/* $Id$ */
#include "caravan.h"

#include "common/aicore/path_finding.h"
#include "common/aicore/pf_tools.h"
#include "log.h"

/***************************************************************************
  Utility functions for parameters.
 ***************************************************************************/
void caravan_parameter_init_default(struct caravan_parameter *parameter) {
  parameter->horizon = FC_INFINITY;
  parameter->ignore_transit_time = FALSE;
  parameter->discount = .95;
  parameter->objective = CARAVAN_OPTIMIZE_BOTH;
  parameter->account_for_broken_routes = FALSE;
  parameter->convert_trade = FALSE;
  parameter->allow_foreign_trade = FALSE;
}

bool caravan_parameter_is_legal(const struct caravan_parameter *parameter) {
  if(parameter->horizon == FC_INFINITY && parameter->discount == 1.0
      && parameter->objective == CARAVAN_OPTIMIZE_BOTH) {
    return FALSE;
  }
  return TRUE;
}

void caravan_parameter_log(const struct caravan_parameter *parameter,
    int loglevel) {
  freelog(loglevel, "parameter {\n"
      "  horizon = %d\n"
      "  discount= %g\n"
      "  objective=%d\n"
      "  ignore-transit = %s\n"
      "  account-broken = %s\n"
      "  convert-trade  = %s\n"
      "  allow-foreign  = %s\n"
      "}\n"
      , parameter->horizon
      , parameter->discount
      , parameter->objective
      , parameter->ignore_transit_time ? "yes" : "no"
      , parameter->account_for_broken_routes ? "yes" : "no"
      , parameter->convert_trade ? "yes" : "no"
      , parameter->allow_foreign_trade ? "yes" : "no"
      );
}


/***************************************************************************
  We use the path finding in several places.
  This provides a single implementation of that.  It is critical that
  this function be re-entrant since we call it recursively.

  The callback should return TRUE if it wants to stop searching,
  FALSE otherwise.
 ***************************************************************************/
typedef bool (*search_callback) (void *data,
    const struct city *pcity,
    int arrival_time);

static void caravan_search_from(
    const struct unit *caravan,
    const struct caravan_parameter *param,
    int start_x, int start_y,
    int turns_before,
    search_callback callback,
    void *callback_data) {
  struct pf_map *map;
  struct pf_parameter pfparam;
  int end_time;

  end_time = param->horizon - turns_before;

  pft_fill_unit_parameter(&pfparam, caravan);
  pfparam.start_x = start_x;
  pfparam.start_y = start_y;
  map = pf_create_map(&pfparam);

  /* for every tile in distance order:
     quit if we've exceeded the maximum number of turns
     otherwise, run the callback if we're on a city. */
  while(pf_next(map)) {
    struct pf_position p;
    struct city *pcity;

    pf_next_get_position(map, &p);
    if(p.turn > end_time) {
      break;
    }

    pcity = map_get_city(p.x, p.y);
    if(pcity) {
      bool stop = callback(callback_data, pcity, turns_before + p.turn);
      if(stop) break;
    }
  }

  pf_destroy_map(map);
}




/***************************************************************************
  When the caravan arrives, compute the benefit from the immediate windfall,
  taking into account the parameter's objective.
 ***************************************************************************/
static double windfall_benefit(
    const struct city *src,
    const struct city *dest,
    const struct caravan_parameter *param) {
  if(param->objective == CARAVAN_OPTIMIZE_TRADE) {
    return 0;
  }
  else {
    int bonus = get_caravan_enter_city_trade_bonus(src, dest);
    bool can_establish = can_establish_trade_route(src, dest);

    /* we get the full bonus only if this is a new trade route.
     * Really, g_c_e_c_t_b should compute this.  I copy from unithand.c */
    if(!can_establish) {
      bonus += 2;
      bonus /= 3;
    }

    /* bonus goes to both sci and gold. */
    bonus *= 2;
    return bonus;
  }
}


/***************************************************************************
  Compute the change in the per-turn trade.
 ***************************************************************************/

/* how much does the city benefit from the new trade route?
   how much does the former partner lose? */
static int one_city_trade_benefit(const struct city *pcity,
    int playerid, bool countloser, int newtrade) {
  int losttrade = 0;

  /* if the city is owned by someone else, we don't benefit from the
     new trade (but we might still lose from a broken trade route) */
  if(pcity->owner != playerid) {
    newtrade = 0;
  }

  if(city_num_trade_routes(pcity) < NUM_TRADEROUTES) {
    /* if the city can handle this route, we don't break any old routes */
    losttrade = 0;
  } else {
    int slot;
    struct city *losercity;
    int oldtrade = get_city_min_trade_route(pcity, &slot);
    losercity = find_city_by_id(pcity->trade[slot]);
    /* if we own the city, the trade benefit is only by how much
       better we are than the old trade route */
    if(pcity->owner == playerid) {
      newtrade -= oldtrade;
    }

    /* if the city that lost a trade route is one of ours, and if we
       care about accounting for the lost trade, count it. */
    if(countloser && losercity->owner == playerid) {
      losttrade = oldtrade;
    }
  }

  /* find the increase or decrease in trade we benefit from */
  return newtrade - losttrade;
}

/* Compute one_trade_benefit for both cities and do some other logic. */
static double trade_benefit(
    int caravan_owner,
    const struct city *src,
    const struct city *dest,
    const struct caravan_parameter *param) {
  /* do we care about trade at all? */
  if(param->objective == CARAVAN_OPTIMIZE_WINDFALL) {
    return 0;
  }

  /* first, see if a new route is made. */
  if(!can_establish_trade_route(src, dest)) {
    return 0;
  }

  if(!param->convert_trade) {
    bool countloser = param->account_for_broken_routes;
    int newtrade = trade_between_cities(src, dest);

    return one_city_trade_benefit(src, caravan_owner, countloser, newtrade)
      +    one_city_trade_benefit(dest, caravan_owner, countloser, newtrade);
  }
  else {
    die("unimplemented functionality: using CM to calculate trade");
  }
}


/***************************************************************************
  Compute the discounted reward from the trade route, assuming that we
  establish it after 'arrival_time' turns.
 ***************************************************************************/
static double get_discounted_reward(
    const struct city *src,
    const struct city *dest,
    int arrival_time,
    const struct caravan_parameter *parameter) {
  double total=0;
  /* assume the owner of src is the owner of the caravan */
  double trade;
  double windfall;
  int i;

  /* if no trade is allowed between these two, just quit. */
  if(!can_cities_trade(src, dest)) {
    return 0;
  }

  /* if no foreign trade is allowed, just quit. */
  if(!parameter->allow_foreign_trade && src->owner != dest->owner) {
    return 0;
  }

  trade = trade_benefit(src->owner, src, dest, parameter);
  windfall = windfall_benefit(src, dest, parameter);

  /* Compute the discounted contribution from trade.
   * i<horizon is to handle a finite horizion
   * trade>=.1 is to handle an infinite horizon (slightly approximated)
   */
  i = 0;
  while(trade >= .1 && i < parameter->horizon) {
    trade *= parameter->discount;
    if(i > arrival_time) {
      total += trade;
    }
    i++;
  }

  /* Compute the discounted contribution from windfall */
  if(windfall != 0) {
    int i;
    /* If arrival_time is turn 0, then no discount. */
    double winddiscount = 1;
    for(i = 0; i < arrival_time; i++) {
      winddiscount *= parameter->discount;
    }
    total += windfall * winddiscount;
  }

  if(parameter->callback) {
    parameter->callback(src, dest, total, parameter->callback_data);
  }

  return total;
}

/***************************************************************************
  Compute the benefit of moving the caravan to dest.
 ***************************************************************************/

/* ignore transit time */
static double caravan_evaluate_notransit(
    const struct unit *caravan,
    const struct city *dest,
    const struct caravan_parameter *param) {
  const struct city *src = find_city_by_id(caravan->homecity);
  return get_discounted_reward(src, dest, 0, param);
}


/* use caravan_search to take transit time into account */
struct cewt_data {
  const struct city *src;
  const struct city *dest;
  const struct caravan_parameter *param;

  double val;
};

static bool cewt_callback(void *vdata, const struct city *dest,
    int arrival_time) {
  struct cewt_data *data = vdata;
  if(dest == data->dest) {
    data->val = get_discounted_reward(data->src, data->dest,
        arrival_time, data->param);
    return TRUE;
  }
  else {
    return FALSE;
  }
}

static double caravan_evaluate_withtransit(
    const struct unit *caravan,
    const struct city *dest,
    const struct caravan_parameter *param) {
  struct cewt_data data;
  data.src   = find_city_by_id(caravan->homecity);
  data.dest  = dest;
  data.param = param;
  data.val   = 0;
  caravan_search_from(caravan, param, caravan->x, caravan->y, 0,
      cewt_callback, &data);
  return data.val;
}


/* externally-visible function for users to call */
double caravan_evaluate(
    const struct unit *caravan,
    const struct city *dest,
    const struct caravan_parameter *param) {
  if(param->ignore_transit_time) {
    return caravan_evaluate_notransit(caravan, dest, param);
  } else {
    return caravan_evaluate_withtransit(caravan, dest, param);
  }
}


/***************************************************************************
  Find the best destination for the caravan.
 ***************************************************************************/

/* ignoring transit time */
static double caravan_find_best_destination_notransit(
    const struct unit *caravan,
    const struct caravan_parameter *param,
    const struct city **destout) {
  double best = 0;
  const struct city *src = find_city_by_id(caravan->homecity);
  *destout = NULL;
  cities_iterate(dest) {
    double val = get_discounted_reward(src, dest, 0, param);
    if(val > best) {
      best = val;
      *destout= dest;
    }
  } cities_iterate_end;
  return best;
}

/* Use caravan_search to take transit time into account. */
struct cfbdw_data {
  const struct caravan_parameter *param;
  const struct city *src;

  double best;
  const struct city *best_dest;
};

static bool cfbdw_callback(void *vdata, const struct city *dest,
    int arrival_time) {
  struct cfbdw_data *data = vdata;
  double val;

  val = get_discounted_reward(data->src, dest, arrival_time, data->param);
  if(val > data->best) {
    data->best = val;
    data->best_dest = dest;
  }
  return FALSE; /* don't stop. */
}

static double caravan_find_best_destination_withtransit(
    const struct unit *caravan,
    const struct caravan_parameter *param,
    const struct city *src,
    int turns_before,
    const struct city **destout) {
  int start_x, start_y;
  struct cfbdw_data data;
  data.src = src;
  data.param = param;
  data.best = 0;
  data.best_dest = NULL;

  if(src->id != caravan->homecity) {
    start_x = src->x;
    start_y = src->y;
  } else {
    start_x = caravan->x;
    start_y = caravan->y;
  }

  caravan_search_from(caravan, param, start_x, start_y, turns_before,
      cfbdw_callback, &data);
  *destout = data.best_dest;
  return data.best;
}


/*
 * The externally visible function.
 */
double caravan_find_best_destination(
    const struct unit *caravan,
    const struct caravan_parameter *parameter,
    const struct city **destout) {
  if(parameter->ignore_transit_time) {
    return caravan_find_best_destination_notransit(caravan,
        parameter, destout);
  } else {
    return caravan_find_best_destination_withtransit(caravan,
        parameter, find_city_by_id(caravan->homecity), 0, destout);
  }
}



/***************************************************************************
  Find the best pair-wise trade route.
 ***************************************************************************/

/* Ignore transit times. */
static double caravan_optimize_notransit(
    const struct unit *caravan,
    const struct caravan_parameter *param,
    const struct city **srcout,
    const struct city **destout) {
  struct player *player;
  double best = 0;

  player = get_player(caravan->owner);
  city_list_iterate(player->cities, src) {
    cities_iterate(dest) {
      double val = get_discounted_reward(src, dest, 0, param);
      if(val > best) {
        best = val;
        *srcout = src;
        *destout= dest;
      }
    } cities_iterate_end;
  } city_list_iterate_end;
  return best;
}

/* Use caravan_search to take transit times into account.
   Uses caravan_find_best_destination as a subroutine. */
struct cowt_data {
  const struct caravan_parameter *param;
  const struct unit *caravan;

  double best;
  const struct city *best_src;
  const struct city *best_dest;
};

static bool cowt_callback(void *vdata, const struct city *src, 
    int arrival_time) {
  struct cowt_data *data = vdata;
  double val;
  const struct city *dest;

  val = caravan_find_best_destination_withtransit(
      data->caravan, data->param,
      src, arrival_time,
      &dest);

  if(val > data->best) {
    data->best = val;
    data->best_src = src;
    data->best_dest = dest;
  }
  return FALSE; /* don't stop searching */
}

static double caravan_optimize_withtransit(
    const struct unit *caravan,
    const struct caravan_parameter *param,
    const struct city **srcout,
    const struct city **destout) {
  struct cowt_data data;
  data.param = param;
  data.caravan = caravan;
  data.best = 0;
  data.best_src = NULL;
  data.best_dest = NULL;

  caravan_search_from(caravan, param, caravan->x, caravan->y, 0,
      cowt_callback, &data);
  *srcout = data.best_src;
  *destout= data.best_dest;
  return data.best;
}

/* The externally-visible function. */
double caravan_optimize_allpairs(
    const struct unit *caravan,
    const struct caravan_parameter *param,
    const struct city **src,
    const struct city **dest) {
  if(param->ignore_transit_time) {
    return caravan_optimize_notransit(caravan, param, src, dest);
  } else {
    return caravan_optimize_withtransit(caravan, param, src, dest);
  }
}
