#ifndef FC__CARAVAN_H
#define FC__CARAVAN_H

#include "config.h"
#include "shared.h"
struct city;
struct unit;

/**
 * An advisor for using caravans optimally.
 * Set the caravan_parameter, then call caravan_optimize for every destination
 * or source/dest pair we care to optimize.
 * The optimization does not take into account other caravans in transit.
 * It also knows nothing about moving caravans except what pathfinding will
 * tell it -- ferries, for instance, aren't handled.  Set horizon to
 * FC_INFINITY to work around this.
 */

struct caravan_parameter {
    /*
     * How many turns to consider when optimizing.
     */
    int horizon;

    /*
     * Discount factor (that is, 1 / interest rate).
     * The worth of a trade route is equivalent to:
     * sum_{i=0}^{horizon} value in i turns * discount^i
     * The discount should be in (0..1] ;
     * A discount of 1 means future earnings are worth the same as immediate
     * earnings.  A discount of .95 means a dollar next turn is worth only 95
     * cents today; or, in other words, 95 cents in the bank today would yield
     * a dollar next year.
     * If the discount is 1 and the horizon infinite, optimizing for both trade
     * and windfall is non-sensical.
     */
    double discount;

    /*
     * What to consider:
     * - just the immediate windfall when the caravan reaches destination?
     * - just the trade?
     * - both?  Then we take double the windfall + effect of trade (since
     *          windfall goes both to gold and science).
     * Note: if the horizon is infinite, "both" doesn't really make sense,
     * so I treat it as just trade.  Really we'd want to say that any trade
     * increase
     */
    enum {
        CARAVAN_OPTIMIZE_WINDFALL,
        CARAVAN_OPTIMIZE_TRADE,
        CARAVAN_OPTIMIZE_BOTH
    } objective;

    /*
     * Normally, we'd want to compute the time it takes to establish the
     * trade route.
     * There are two reasons to ignore the transit time:
     * (1) it may be infinite (i.e. requires crossing an ocean)
     * (2) it's slow to compute it.
     */
    bool ignore_transit_time;

    /*
     * A new trade route may break an old trade route or two.
     * We should take account of the cost of the broken trade route,
     * but I'm too lazy right now.
     */
    bool account_for_broken_routes;

    /*
     * Should we allow trading with allies?
     * BUG: currently, we allow trading with the enemy; this should never
     * occur.
     */
    bool allow_foreign_trade;

    /*
     * The 'effect of trade' can be computed exactly by an expensive
     * calculation which will take into account buildings, tax rates, etc; or
     * by a cheap calculation that just says that each trade is worth one
     * science or gold.
     */
    bool convert_trade;

    /*
     * This callback, if non-null, is called whenever a trade route
     * is evaluated.  One intended usage is for collecting all the 
     * trade routes into a sorted list.
     */
    void (*callback)(const struct city *src, const struct city *dest,
            double val, void *data);
    void *callback_data;
};

void caravan_parameter_init_default(struct caravan_parameter *parameter);
bool caravan_parameter_is_legal(const struct caravan_parameter *parameter);
void caravan_parameter_log(const struct caravan_parameter *parameter,
        int loglevel);

/**
 * Find the value of establishing a trade route to the dest city
 * using the given caravan.
 */
double caravan_evaluate(
        const struct unit *caravan,
        const struct city *dest,
        const struct caravan_parameter *parameter);

/*
 * Find the best destination for this caravan.
 */
double caravan_find_best_destination(
        const struct unit *caravan,
        const struct caravan_parameter *parameter,
        const struct city **destout);

/*
 * Find the best trade route for this caravan, possibly changing
 * home city.
 */
double caravan_optimize_allpairs(
        const struct unit *caravan,
        const struct caravan_parameter *parameter,
        const struct city **src,
        const struct city **dest);

#endif
